//
// Copyright (c) 2014 Secret Geometry, Inc.  All rights reserved.
//

#ifndef TERMKIT_TERMINAL_HPP
#define TERMKIT_TERMINAL_HPP

#include "Types.hpp"
#include "Config.hpp"
#include <vector>
#include <map>

namespace term {

//
// Terminal
//
// The top level, cross-platform C++ class for TermKit.
// Creates and hooks up a Connection (PTY or SSH), a Parser, a Dispatcher, a Context, and a TextArea.
// Handles input, and holds output.
// Does work in calls to update().
//

template
<
    typename ConnectionClass,
    typename KeyboardClass,
    typename TextAreaClass,
    typename ContextClass,
    typename DispatcherClass,
    typename ParserClass
>
class Terminal
{
  public:
    
    // 
    Terminal(const char* server,
             const char* user,
             const char* password,
             unsigned int port,
             size_t width,
             size_t height,
             size_t maxLines,
             BeepCallback beepFunc,
             ResizeCallback resizeFunc,
             const char* language,
             const char* termIdentifier,
             const char* hostAppName,
             StringMap& otherEnvars)
    :
        _connection(server, user, password, port, width, height, language, termIdentifier, hostAppName, otherEnvars),
        _keyboard(_connection),
        _textArea(width, height, maxLines),
        _context(_textArea, beepFunc, resizeFunc),
        _dispatcher(_context),
        _parser(_dispatcher),
        _readBuffer(8 * 1024 * 1024),
        _hasChanges(false)
    {
        #ifdef DO_SANITY_CHECKS
            _context.sanityCheck();
        #endif
    }
    
    ~Terminal()
    {
        // Intentionally empty.  Some Obj-C++ compilers complain if this isn't explicitly defined.
    }
    
    void start()
    {
        _connection.start();
    }
    
    void logout()
    {
        _connection.logout();
    }
    
    void suspend()
    {
        _connection.suspend();
    }
    
    void resume()
    {
        LineLocation lastLineLoc = _context.textArea().lastLineEnd();
        GridLocation lastGridLoc = _context.textArea().gridLocForLineLoc(lastLineLoc);
        
        std::wstring message = L"Resuming...";
        
        _context.setCursorLoc(lastGridLoc);
        _context.pushCursorToFirstColOfNewLine();
        _context.echo(message.data(), message.size());
        _context.pushCursorToFirstColOfNewLine();
        
        _hasChanges = true;
        
        _connection.resume();
    }
    
    ConnectionClass& connection()
    {
        return _connection;
    }
    
    KeyboardClass& keyboard()
    {
        return _keyboard;
    }
    
    ContextClass& context()
    {
        return _context;
    }
    
    const ContextClass& context() const
    {
        return _context;
    }
    
    TextAreaClass& textArea()
    {
        return _textArea;
    }
    
    const TextAreaClass& textArea() const
    {
        return _textArea;
    }
    
    size_t columnCount() const
    {
        return _textArea.width();
    }
    
    size_t rowCount() const
    {
        return _textArea.height();
    }
    
    void setColumnCount(size_t cols)
    {
        resize(cols, rowCount());
    }
    
    void setRowCount(size_t rows)
    {
        resize(columnCount(), rows);
    }
    
    void resize(size_t cols, size_t rows)
    {
        #ifdef DO_SANITY_CHECKS
            _context.sanityCheck();
        #endif
        
        update();
        
        #ifdef DO_SANITY_CHECKS
            _context.sanityCheck();
        #endif
        
        _context.resize(cols, rows);
        _connection.resize(cols, rows);
        _hasChanges = true;
        
        #ifdef DO_SANITY_CHECKS
            _context.sanityCheck();
        #endif
        
        update();
        
        #ifdef DO_SANITY_CHECKS
            _context.sanityCheck();
        #endif
    }
    
    GridLocation cursorLoc() const
    {
        return _context.cursorLoc();
    }
    
    const std::string& activeProcessName() const
    {
        return _connection.activeProcessName();
    }
    
    const std::vector<std::string>& allProcessNames() const
    {
        return _connection.allProcessNames();
    }
    
    bool hasChanges() const
    {
        return _hasChanges;
    }
    
    void setHasChanges(bool b)
    {
        _hasChanges = b;
    }
    
    ConnectionState state() const
    {
        return _connection.state();
    }
    
    const std::string& message() const
    {
        return _connection.message();
    }
    
    void clearErrorMessage()
    {
        _connection.clearErrorMessage();
    }
    
    void clearBuffer()
    {
        _context.clearBuffer();
        _hasChanges = true;
    }
    
    void write(uint8_t c)
    {
        write(&c, 1);
    }
    
    void write(const char* s)
    {
        if (s == nullptr)
            return;
        
        size_t length = strlen(s);
        
        write((const uint8_t*)s, length);
    }
    
    void write(const uint8_t* str, size_t count)
    {
        size_t bytesWritten = 0;
        size_t bytesRemaining = count;
        
        while (true)
        {
            ssize_t writeCount = _connection.write(str, bytesRemaining);
            
            if (writeCount < 0)
                break;
            
            str += writeCount;
            bytesWritten += writeCount;
            bytesRemaining -= writeCount;
            
            if (bytesWritten < count)
                update();
            else
                break;
        }
    }
    
    void update()
    {
        update(_readBuffer.size());
    }
    
    void update(size_t numBytes)
    {
        // Handy for debugging.
        //numBytes = 1;
        
        _connection.update();
        
        if (_connection.state() == ConnectionState_Disconnected ||
            _connection.state() == ConnectionState_Suspended)
            return;
        
        if (numBytes == 0)
            return;
        
        if (numBytes > _readBuffer.size())
            numBytes = _readBuffer.size();
        
        // Grab any new data from the connection and put it into the buffer.
        ssize_t count = _connection.read(_readBuffer.data(), numBytes);
        
        if (count <= 0)
            return;
        
        #ifdef DO_SANITY_CHECKS
            _context.sanityCheck();
        #endif
        
        // Pass the buffer of new data to the parser.  This starts a complex chain of actions.
        // The parser breaks the data into text and various types of command sequences; it sends these to the dispatcher.
        // The dispatcher translates everything it receives into calls to member functions of the context.
        // The context's methods modify the text, cursor, style, scroll, etc.
        _parser.parse(_readBuffer.data(), count);
        
        // If the context has prepared a response, write it back to the connection.
        write(_context.getResponse());
        
        _hasChanges = true;
        
        _connection.updateProcessInfo();
        
        #ifdef DO_SANITY_CHECKS
            _context.sanityCheck();
        #endif
        
        return;
    }
    
  private:
    
    ConnectionClass _connection;
    KeyboardClass   _keyboard;
    TextAreaClass   _textArea;
    ContextClass    _context;
    DispatcherClass _dispatcher;
    ParserClass     _parser;
    UInt8Vector     _readBuffer;
    bool            _hasChanges;
};

}

#endif
